import unittest
import numpy as np
from itertools import chain, combinations

from classifier.mediator.components import helper as h


class TestHelper(unittest.TestCase):

    def test_powerset(self):
        s = {2, 4, 6, 8}

        powerset1 = h.get_powerset(s, 4)
        expected_powerset1 = [frozenset(), frozenset([2]), frozenset([4]), frozenset([6]), frozenset([8]),
                              frozenset([2, 4]), frozenset([2, 6]), frozenset([2, 8]), frozenset([4, 6]),
                              frozenset([4, 8]), frozenset([6, 8]), frozenset([2, 4, 6]), frozenset([2, 4, 8]),
                              frozenset([2, 6, 8]), frozenset([4, 6, 8]), frozenset([2, 4, 6, 8])]

        self.assertCountEqual(list(powerset1), expected_powerset1)

        # test additivity powerset
        powerset2 = h.get_powerset(s, additivity=2)
        expected_powerset2 = [frozenset(), frozenset([2]), frozenset([4]), frozenset([6]), frozenset([8]),
                              frozenset([2, 4]), frozenset([2, 6]), frozenset([2, 8]), frozenset([4, 6]),
                              frozenset([4, 8]), frozenset([6, 8])]

        self.assertCountEqual(list(powerset2), expected_powerset2)

    def test_powerset_dictionary(self):
        s = {1, 2, 3, 4}
        additivity = 2

        powerset_dict = h.get_powerset_dictionary(s, additivity)
        expected_powerset = chain.from_iterable(combinations(s, r) for r in range(1, additivity + 1))

        expected_powerset_list = []
        for item in expected_powerset:
            expected_powerset_list.append(frozenset(item))

        for key, item in powerset_dict.items():
            self.assertEqual(powerset_dict[key], expected_powerset_list[key - 1])

    def test_additivity_powerset(self):
        s = [1, 2, 3]
        additivity = 2

        powerset = h._get_additivity_powerset(s, additivity)
        expected_powerset = [frozenset(), frozenset({1}), frozenset({2}), frozenset({3}),
                             frozenset({1, 2}), frozenset({1, 3}), frozenset({2, 3})]

        self.assertListEqual(list(powerset), expected_powerset)

    def test_subset_dictionary_list(self):
        s = {1, 2, 3, 4}
        additivity = 3

        subset_dict_list = h.get_subset_dictionary_list(s, additivity)
        expected_dict_list = [{1: frozenset({1}), 2: frozenset({2}), 5: frozenset({1, 2})},
                              {1: frozenset({1}), 3: frozenset({3}), 6: frozenset({1, 3})},
                              {1: frozenset({1}), 4: frozenset({4}), 7: frozenset({1, 4})},
                              {2: frozenset({2}), 3: frozenset({3}), 8: frozenset({2, 3})},
                              {2: frozenset({2}), 4: frozenset({4}), 9: frozenset({2, 4})},
                              {3: frozenset({3}), 4: frozenset({4}), 10: frozenset({3, 4})},
                              {1: frozenset({1}), 2: frozenset({2}), 3: frozenset({3}), 5: frozenset({1, 2}), 6: frozenset({1, 3}), 8: frozenset({2, 3}), 11: frozenset({1, 2, 3})},
                              {1: frozenset({1}), 2: frozenset({2}), 4: frozenset({4}), 5: frozenset({1, 2}), 7: frozenset({1, 4}), 9: frozenset({2, 4}), 12: frozenset({1, 2, 4})},
                              {1: frozenset({1}), 3: frozenset({3}), 4: frozenset({4}), 6: frozenset({1, 3}), 7: frozenset({1, 4}), 10: frozenset({3, 4}), 13: frozenset({1, 3, 4})},
                              {2: frozenset({2}), 3: frozenset({3}), 4: frozenset({4}), 8: frozenset({2, 3}), 9: frozenset({2, 4}), 10: frozenset({3, 4}), 14: frozenset({2, 3, 4})}]

        self.assertListEqual(subset_dict_list, expected_dict_list)

    def test_get_feature_subset(self):
        array_1 = np.array([4, 21, 1, 6, 9])
        sorted_array_1 = np.sort(array_1)

        permutation_1 = h._get_permutation_position(array_1, sorted_array_1)

        subset_1_index_1 = h.get_feature_subset(array_1, 1)
        expected_subset_1 = {3, 1, 4, 5, 2}
        self.assertSetEqual(subset_1_index_1, expected_subset_1)

        subset_2_index_2 = h.get_feature_subset(array_1, 2)
        expected_subset_2 = {1, 4, 5, 2}
        self.assertSetEqual(subset_2_index_2, expected_subset_2)

    def test_get_permutation_position(self):
        array_1 = np.array([4, 21, 1, 6, 9])
        sorted_array_1 = np.sort(array_1)

        expected_permutation_1 = {1: 3, 2: 1, 3: 4, 4: 5, 5: 2}
        permutation_1 = h._get_permutation_position(array_1, sorted_array_1)
        self.assertDictEqual(permutation_1, expected_permutation_1)


if __name__ == '__main__':
    unittest.main()



import numpy as np

import unittest
from classifier.mediator.components.choquet_integral import ChoquetIntegral


class TestChoquetIntegral(unittest.TestCase):
    def test_compute_aggregation_value(self):
        X = [[0.1, 0.2, 0.3], [0.2, 0.3, 0.4], [0.3, 0.4, 0.5], [0.4, 0.5, 0.6]]
        y = [0, 0, 0, 1]

        additivity = 3
        number_of_instances = np.shape(X)[0]

        moebius_coefficients = [0.32621127227372915, 0.1096778349966972, 0.14071448801729294, 0.2181001452983398,
                                0.06220493806514601, 0.1071507113468563, 0.035940610001938655]

        capacities = [0.32621127227372915, 0.1096778349966972, 0.14071448801729294, 0.6539892525687662,
                      0.5291306983561681, 0.35754303436084645, 1.0]

        # Calculation of integration values and counter-checking from http://www.isc.senshu-u.ac.jp/~thc0456/Efuzzyweb/
        # by Eiichiro Takahagi, Senshu University, Kawasaki, Japan
        integrated_values = [0.149826, 0.249826, 0.349826, 0.449826]

        choquet_integral = ChoquetIntegral(additivity, moebius_coefficients)
        for i in range(number_of_instances):
            self.assertAlmostEqual(choquet_integral.compute_utility_value(X[i]), integrated_values[i], places=6)

    def test_feature_minima(self):
        X = [[0.1, 0.2, 0.3], [0.2, 0.3, 0.4], [0.3, 0.4, 0.5], [0.4, 0.5, 0.6]]
        y = [0, 0, 0, 1]

        additivity = 3
        number_of_instances = np.shape(X)[0]

        moebius_coefficients = [0.32621127227372915, 0.1096778349966972, 0.14071448801729294, 0.2181001452983398,
                                0.06220493806514601, 0.1071507113468563, 0.035940610001938655]

        choquet_integral = ChoquetIntegral(additivity, moebius_coefficients)

        for i in range(number_of_instances):
            x = X[i]
            expected_minima_dict = {1: x[0], 2: x[1], 3: x[2], 4: min(x[0], x[1]), 5: min(x[0], x[2]),
                                    6: min(x[1], x[2]), 7: min(x[0], x[1], x[2])}
            actual_minima_dict = choquet_integral.feature_minima_of_instance(x)
            self.assertDictEqual(expected_minima_dict, actual_minima_dict)


if __name__ == '__main__':
    unittest.main()



import unittest

from sklearn.utils.estimator_checks import check_estimator
from classifier.choquet_classifier import ChoquetClassifier


class TestChoquetClassifier(unittest.TestCase):
    def test_compatibility(self):
        cc = ChoquetClassifier()
        check_estimator(cc)

    def test_documentation(self):
        X = [[1, 3, 2], [1, 0, 3]]
        y = [1, 0]

        # test data
        Z = [[1, 1, 2], [2, 1, 3]]

        # first example
        cc = ChoquetClassifier()
        cc.fit(X, y)
        result = cc.predict(Z)

        self.assertListEqual([0, 0], list(result))

        # second example
        cc = ChoquetClassifier(additivity=3, regularization=1)
        cc.fit(X, y)
        result = cc.predict(Z)

        self.assertListEqual([0, 1], list(result))

        # third example
        y = [2, 1]
        cc = ChoquetClassifier()
        cc.fit(X, y)
        result = cc.predict(Z)

        self.assertListEqual([1, 1], list(result))

        y = ['two', 'one']
        cc = ChoquetClassifier()
        cc.fit(X, y)
        result = cc.predict(Z)

        self.assertListEqual(['one', 'one'], list(result))


if __name__ == '__main__':
    unittest.main()



import numpy as np
from itertools import chain, combinations


def get_feature_subset(x, index):
    """Get a subset A_i={p(c_i), ..., p(c_m)} with a permutation p"""

    m = len(x)
    sorted_x = np.sort(x)
    permutation = _get_permutation_position(x, sorted_x)

    result = [permutation[i] for i in range(index, m + 1)]
    return frozenset(result)


def get_powerset(s, additivity):
    result = list()
    items = s
    if additivity is not None:
        powerset = _get_additivity_powerset(items, additivity)
    else:
        powerset = chain.from_iterable(combinations(items, r) for r in range(len(items) + 1))

    for item in powerset:
        result.append(frozenset(item))

    return np.array(result)


def get_powerset_dictionary(s, additivity):
    powerset = get_powerset(s, additivity)
    dict_powerset = dict(enumerate(powerset[1:].flatten(), 1))
    return dict_powerset


def get_subset_dictionary_list(s, additivity):
    powerset_dict = get_powerset_dictionary(s, additivity)
    dict_list = list()

    if additivity == 1:
        subset_list = list(powerset_dict.values())
    else:
        subset_list = list(powerset_dict.values())[len(s):]

    additivity_subset_list = [x for x in subset_list if len(x) <= additivity]

    for i in additivity_subset_list:
        dict_list.append({key: val for key, val in powerset_dict.items() if val <= i})

    return dict_list


def _get_additivity_powerset(items, additivity):
    result = list()

    if len(items) < additivity:
        raise Exception("additivity must be less than or equal to number of features")

    powerset = chain.from_iterable(combinations(items, r) for r in range(additivity + 1))

    for item in powerset:
        result.append(frozenset(item))

    return np.array(result)


def _get_permutation_position(x, sorted_x):
    result = {}

    # to cover duplicate values, create array instead of pure values
    for i in range(len(x)):
        result[i + 1] = np.nonzero(x == sorted_x[i])[0][0] + 1

    return result



import numpy as np
from . import helper as h


class ChoquetIntegral:
    """Choquet integral

    Class to compute the Choquet utility value and
    store the moebius coefficients of the fuzzy measure.

    Parameters
    -------
    additivity : int
            Additivity of fuzzy measure.

    moebius_coefficients : list
            List of computed moebius coefficients, with the same internal
            ordering as the moebius_transform dictionary, calculated in
            the parameter estimation, since in Python 3.7+ dictionaries
            are guaranteed to be insertion-ordered. This insertion-order
            is used to easily calculate the terms which, when summed up,
            give the Integral value.
    """

    def __init__(self, additivity, moebius_coefficients):
        self.additivity = additivity
        self.moebius_coefficients = moebius_coefficients

    def compute_utility_value(self, instance):
        """Compute the utility value for a given instance

        Parameters
        -------
        instance:  array-like of shape (1, n_features)
            Instance, where n_features is the number of features.

        Returns
        -------
        choquet_value : float
            Integral value computed by the Choquet Integral.
        """

        choquet_value = 0
        feature_minimum_dictionary = self.feature_minima_of_instance(instance)
        for j in range(len(self.moebius_coefficients)):

            choquet_value += self.moebius_coefficients[j] * feature_minimum_dictionary[j + 1]

        return choquet_value

    def feature_minima_of_instance(self, instance):
        """Get minimal feature of every subset and store them
        in a dictionary.


        Parameters
        -------
        instance : array-like of (1, n_features)
            Instance, where n_features is the number of features.

        Returns
        -------
        minima_dict: dict
            Dictionary, containing all insertion-ordered number of subsets as keys, and their
            corresponding minimal feature of instance. Dictionary
            has the same insertion order as the moebius_coefficients dictionary,
            to guarantee an indexed-controlled calculation, i.e.
            the value at key [n+1] in minima_dict corresponds
            to the nth value in moebius_coefficients (the value is the minima
            of all the features represented by the nth moebius_coefficient)
        """

        features = list(range(1, len(instance) + 1))

        powerset_dict = h.get_powerset_dictionary(features, self.additivity)
        minima_dict = {key: np.amin([instance[i - 1] for i in value]) for key, value in powerset_dict.items()}
        return minima_dict



import numpy as np

from .mediator.mediator import Mediator

from sklearn.base import BaseEstimator
from sklearn.base import ClassifierMixin
from sklearn.utils.validation import check_is_fitted


class ChoquetClassifier(BaseEstimator, ClassifierMixin):
    """The Choquet classifier

    Implementation of the Choquet classifier, which was introduced in
    "Learning monotone nonlinear models using the Choquet integral", using the Moebius transform
    presentation of the underlying fuzzy measure.
    This classifier is compatible to scikit-learn, i.e. it can be used like any scikit-learn estimator.

    Parameters
    -------
    additivity : int, default=1
        The additivity of the underlying fuzzy measure. Additivity takes interaction between features into account,
        i.e. the additivity value determines the maximum number of interacting features and hence the maximum
        number of moebius coefficients to be estimated. More precisely, with a k-additive measure, sum_i (n over i),
        with n = n_features, coefficients. The value 1 represents a simple additive measure with no interaction
        between features. The default value 2 represents a simple interaction between 2 features.

    regularization: in, default=None
        the regularization parameter of the L1-Regularization in the parameter estimation. Determines the strength of
        regularization of the fitting process.
    """

    def __init__(self, additivity=2, regularization=None):
        self.additivity = additivity
        self.regularization = regularization

    def fit(self, X, y):
        """Initialize the parameters of the Choquet classifier

        Initialize the feature transformation and the parameters: threshold, scaling and moebius transform of
        the capacity, with the initialized hyperparameter for a given dataset.

        Parameters
        ------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features. The number of features
            has to be more or equal to the additivity.

        y : array-like of shape (n_samples,)
            Target labels to X.

        Returns
        -------
        self: ChoquetClassifier
            Fitted estimator.
        """
        self.mediator_ = Mediator()

        X, y = self.mediator_.check_train_data(X, y)

        self.classes_, y = np.unique(y, return_inverse=True)

        self.mediator_.fit_components(X, y, self.additivity, self.regularization)

        self.n_features_in_ = self.mediator_.number_of_features

        return self

    def predict(self, X):
        """Predict classes for X

        Predict the class labels for all samples in X, which were
        specified in fit. The number of features have to match the
        number of features from the train data.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        ------
        result : array-like of shape (n_samples,)
                The predicted classes.
        """
        check_is_fitted(self)

        X = self.mediator_.check_test_data(X)

        result = self.mediator_.predict_classes(X)

        return self.classes_[result]

    # ===============================================================
    # Functions for compatibility with scikit-learn. Not for general usage
    # ===============================================================

    def get_params(self, deep=True):
        return {'additivity': self.additivity,
                'regularization': self.regularization}

    def _more_tags(self):
        return {'binary_only': True, 'poor_score': True, "no_validation": True,
                "_xfail_checks": {"check_classifiers_one_label": "Model does not work on one label yet"}}




import numpy as np

from .fitter import Fitter
from .predictor import Predictor

from sklearn.utils.validation import check_X_y
from sklearn.utils.validation import check_array


class Mediator:
    """Mediator

    Class to handle estimator and predictor class. Checks the input data, computes the parameters
    of the Choquet classifier and does the classification.
    """

    def __init__(self):
        self.number_of_features = 0
        self.scaling = None

    def check_train_data(self, X, y):
        """Check input data for training

        Compare 'check_X_y" in the documentation of scikit-learn for more information

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like of shape (1, n_samples)
            Target labels to X.

        Returns
        -------
        X : array-like of shape (n_samples, n_features)
            Formatted input data, where n_samples is the number of samples and
            n_features is the number of features.
        y : array-like of shape (1, n_samples)
            Formatted target labels to X.
        """

        X, y = check_X_y(X, y)

        if not self._check_for_regression_targets(y):
            raise ValueError('Unknown label type: ', y)

        return X, y

    def check_test_data(self, X):
        """Check the input data for classification

        Compare 'check_array' in scikit-learn documentation for more information

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.
        Returns
        -------
        X : array-like of shape (n_samples, n_features)
            Formatted input data, where n_samples is the number of samples and
            n_features is the number of features.

        Raises
        -------
        ValueError, if a check should fail.
        """

        X = check_array(X)

        if np.shape(X)[1] != self.number_of_features:
            raise ValueError('Input data does not match number of features')

        return X

    def fit_components(self, X, y, additivity, regularization_parameter):
        """Fit the feature transformation and the parameters: threshold, scaling and moebius transform of
        the capacity for a given input data.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples
            and n_features is the number of features.
        y : array-like of shape (n_samples,)
            Target labels to X.

        additivity : int
            Additivity(Hyperparameter of the Choquet classifier).

        regularization_parameter : float or None
            the regularization parameter of the L1-Regularization
            in the parameter estimation. If None, regularization
            will be set to 1 to estimate the parameters.

        Returns
        -------
        True, indicating that all parameters have been computed.
        """

        self.number_of_features = np.shape(X)[1]
        self.additivity = additivity

        if additivity is not None and self.number_of_features < additivity:
            raise ValueError('Additivity is greater than number of features')

        fitter = Fitter()

        self.feature_transformation = fitter.fit_feature_transformation(X)

        normalized_X = self._get_normalized_X(X, self.feature_transformation)

        self.parameters = fitter.fit_parameters(normalized_X, y, additivity, regularization_parameter)

        self.scaling = fitter.get_scaling_factor()
        self.threshold = fitter.get_threshold()

        self.moebius_transform = fitter.get_moebius_transform_of_capacity()

        return True

    def predict_classes(self, X):
        """Predict classes for input data.

        Predict classes for input data. The function fit_components
        had to be called in advance to provide the feature transformation
        and the parameters.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        result : ndarray of shape (1, n_samples)
            Array containing the classes for the corresponding examples.
        """

        predictor = Predictor()

        normalized_X = self._get_normalized_X(X, self.feature_transformation)

        result = predictor.get_classes(normalized_X, self.additivity, self.parameters)

        return result.ravel()

    def _check_for_regression_targets(self, y):
        """Check target data for regression targets.

        Regression targets are considered to be real non integer numbers.
        This check is necessary to be compatible with scikit-learn.
        This check is copied from the Sugeno classifier by Sven Meyer:
        https://github.com/smeyer198/sugeno-classifier/blob/main/classifier/mediator/mediator.py

        Parameters
        -------
        y : array-like of shape (1,)
            Target labels.

        Returns
        -------
        False, if there is a regression target, true otherwise.
        """

        for value in y:
            # check for numeric value
            if not (isinstance(value, (int, float))
                    and not isinstance(value, bool)):
                continue

            if not float(value).is_integer():
                return False

        return True

    def _print_moebius_transform(self):
        moebius_coefficient = self.parameters[2:]
        for key, value in moebius_coefficient.items():
            print(key, '->', value)

    def _get_normalized_X(self, X, f):
        """Normalize the input data using a Feature Transformation.
        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.
        f : FeatureTransformation
            Feature Transformation.
        Returns
        -------
        result : ndarray of shape (n_samples, n_features)
            Normalized input data, where n_samples is the number of samples
            and n_features the number of features.
        """
        return f(X)



import numpy as np

from .choquet_integral import ChoquetIntegral


class ChoquisticRegression:
    """Choquistic Regression

    This class implements the Choquistic Regression calculation
    and stores all estimated parameters

    Parameters
    -------
    additivity : int
            Additivity of fuzzy measure.

    parameters : dict
            Dictionary, describing all fitted parameters (scaling, threshold,
            moebius coefficients) in the following manner:
            {'gamma': scaling, 'beta': threshold, frozenset({1}): m({1}),...}
    """

    def __init__(self, additivity, parameters):
        self.additivity = additivity
        self.scaling = parameters['gamma']
        self.threshold = parameters['beta']
        self.moebius_coefficients = list(parameters.values())[2:]

    def compute_regression_value(self, instance):
        """Computes the regression value for given instance
         with fitted parameters.

        Parameters
        -------
        instance:  array-like of shape (1, n_features)
            Instance, where n_features is the number of features.

        Returns
        -------
        regression value: float
            Probability of positive class, computed by the choquistic
            regression model.
        """

        choquet_integral = ChoquetIntegral(self.additivity, self.moebius_coefficients)

        gamma = self.scaling
        beta = self.threshold

        utility_value = choquet_integral.compute_utility_value(instance)

        regression_value = 1 / (1 + np.exp(-gamma * (utility_value - beta)))

        return regression_value



import numpy as np

from .components.choquistic_regression import ChoquisticRegression


class Predictor:
    """Predictor

    This class implements the methods for the classification.
    """

    def __init__(self):
        pass

    def get_classes(self, X, additivity, parameters):
        """Get classes for input data.

        Get the classes for given input data. Uses the computed parameters
        (scaling, threshold, moebius coefficients)

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        additivity : int
            Additivity of fuzzy measure.

        parameters : dict
            Dictionary, describing all fitted parameters (scaling, threshold,
            moebius coefficients) in the following manner:
            {'gamma': scaling, 'beta': threshold, frozenset({1}): m({1}),...}

        Returns
        -------
        result: ndarray of shape (1, n_samples)
            Array containing the classes for the corresponding examples.
        """

        choquistic_regression = ChoquisticRegression(additivity, parameters)

        result = self._get_classes_for_X(X, choquistic_regression)

        return result

    def _get_classes_for_X(self, X, regression_model):
        """Get classes for input data.

        Get classes for input data. Use the regression model
        for the classification.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        regression_model : ChoquisticRegression
            The Choquistic Regression class.

        Returns
        -------
        result : ndarray of shape (1, n_samples)
            Array containing the classes for the corresponding examples.
        """

        result = list()
        for x in X:
            regression_value = regression_model.compute_regression_value(x)
            cl = self._get_decision(regression_value)

            result.append(cl)

        return np.array(result)

    def _get_decision(self, regression_value):
        """Get class for regression value

        Parameters
        -------
        regression_value : float
            calculated regression value
        """

        if regression_value >= 0.5:
            return np.array([1])
        else:
            return np.array([0])



import numpy as np
from scipy import optimize as opt
from math import comb

from . import helper as h
from .choquet_integral import ChoquetIntegral


class ParameterEstimation:
    """Class to solve the optimization problem for given
    data set X, y.

    Parameters
    -------
    X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features. The number of features
            has to be more or equal to the additivity.

    y : array-like of shape (n_samples,)
            Target labels to X.

    additivity : int, default=1
            The additivity of the underlying fuzzy measure. Additivity takes interaction
            between features into account, i.e. the additivity value determines the maximum
            number of interacting features and hence the maximum number of moebius coefficients
            to be estimated. More precisely, with a k-additive measure, sum_i (n over i),
            with n = n_features, coefficients. The default value 1 represents
            a simple additive measure with no interaction between features.

    regularization_parameter: in, default=None
            the regularization parameter of the L1-Regularization in the parameter estimation.
            Determines the strength of regularization of the fitting process.
    """

    def __init__(self, X, y, additivity, regularization_parameter=None):
        self.X = X
        self.y = y

        self.additivity = additivity

        self.regularization_parameter = regularization_parameter

        self.number_of_features = np.shape(self.X)[1]
        self.number_of_instances = np.shape(self.X)[0]

        # gamma, beta
        self.boundary_constraint = np.array([0, 0])

    def _log_likelihood_function(self, parameters):
        """Log-Likelihood function which is subject to minimization.

        Takes a parameter array and calculates the utility value
        of every instance x of given data set X and employs
        all parameters, instances and corresponding labels
        into the log-likelihood function.

        Parameters
        -------
        parameters: array-like of shape, (1, n_parameters)
            Array containing all parameters to be optimized.
            Order: [scaling, threshold, [moebius_coefficients]]

        Returns
        -------
        result: float
            The calculated value for the log-likelihood.
        """

        gamma = parameters[0]
        beta = parameters[1]
        moebius_coefficients = parameters[2:]

        choquet = ChoquetIntegral(self.additivity, moebius_coefficients)

        result = 0

        for i in range(self.number_of_instances):
            x = self.X[i]
            y = self.y[i]

            choquet_value = choquet.compute_utility_value(x)

            result += gamma * (1 - y) * (choquet_value - beta) + np.log(1 + np.exp(-gamma * (choquet_value - beta)))

        result += self._l1_regularization(moebius_coefficients)

        return result

    def compute_parameters(self):
        """Function to solve the minimization problem, using the scipy-package.

        Uses the constructed bounds and linear constraints to build up
        the whole optimization problem and sets a starting vector x0.
        The minimize method of the optimize class of scipy is then
        called with the log-likelihood, starting vector x0 and the bounds
        and linear constraints. used method is "trust-constr"

        Returns
        ----------
        function: dict
            Containing keys corresponding to parameters: gamma, beta and subsets of [m]. The values are the
            computed values

        Notes
        ----------
        function has form of: function = {g: x, b: y, 1: m({1}), 2: m({2}),..., }
        """

        bounds, constraints = self._set_constraints()

        # set random normalized starting vector
        x0 = np.concatenate(([1], np.ones(1 + self._get_number_of_moebius_coefficients())), axis=0)
        x0 = x0 / np.sum(x0)

        result = opt.minimize(self._log_likelihood_function, x0, options={'verbose': 1}, bounds=bounds,
                              method='trust-constr', constraints=constraints)

        set_list = h.get_powerset_dictionary(list(range(1, self.number_of_features + 1)), self.additivity)
        parameter_dict = {'gamma': result.x[0], 'beta': result.x[1]}

        moebius_results = result.x[2:]
        moebius_dict = {subset: value for key1, subset in set_list.items()
                        for value in moebius_results if key1 == np.where(moebius_results == value)[0] + 1}

        parameter_dict.update(moebius_dict)
        return parameter_dict

    def _set_constraints(self):
        """ Set up the bounds and linear constraints for the optimization problem
            using the Bounds and Constraints classes from scipy.optimize.

        Returns
        -------
        bounds, constraints: Bounds, LinearConstraints
            Returns the datatypes Bounds and LinearConstraints,
            used for the optimization problem.
        """

        number_of_moebius_coefficients = self._get_number_of_moebius_coefficients()

        # set the bounds - order is: gamma, threshold, m(T_1),...,m(T_n) with n being
        lower_bound = [0.1, 0]
        upper_bound = [np.inf, 1]

        for i in range(number_of_moebius_coefficients):
            lower_bound.append(0)
            upper_bound.append(1)

        bounds = opt.Bounds(lower_bound, upper_bound)

        linear_constraint_matrix = self._get_linear_constraint_matrix()

        lower_boundary_limit, upper_boundary_limit = ([1], [1])
        lower_limits = np.concatenate((lower_boundary_limit, np.zeros(linear_constraint_matrix.shape[0] - 1)), axis=0)
        upper_limits = np.concatenate((upper_boundary_limit, np.ones(linear_constraint_matrix.shape[0] - 1)), axis=0)

        lower_limits = lower_limits.flatten()
        upper_limits = upper_limits.flatten()

        linear_constraint = opt.LinearConstraint(linear_constraint_matrix, lower_limits, upper_limits)

        return bounds, linear_constraint

    def get_monotonicity_matrix(self):
        """Set the needed monotonicity constraints for all
        moebius coefficients

        Create all monotonicity constraints for the optimization
        problem and aggregate them in a matrix.

        Returns
        -------
        monotonicity_matrix: array-like of shape (n_monotonicity_constraints, n_moebius_coefficients)
            The matrix representing all monotonicity constraints
            for moebius coefficients.
        """

        number_of_moebius_coefficients = self._get_number_of_moebius_coefficients()
        set_list = h.get_subset_dictionary_list(list(range(1, self.number_of_features + 1)), self.additivity)
        matrix = []
        for j in set_list:
            max_subset = list(j.values())[-1]
            for criterion in max_subset:
                arr = np.zeros(number_of_moebius_coefficients)
                intersection_sets = [key for key, value in j.items() if len(value.intersection({criterion})) > 0]
                monotonicity_array = [1 if index + 1 in intersection_sets else 0 for index, x in enumerate(arr)]
                matrix.append(monotonicity_array)

        monotonicity_matrix = np.array(matrix)

        return monotonicity_matrix

    def _get_linear_constraint_matrix(self):
        """Set the linear constraint matrix needed for the scipy-optimize
        package.

        Use monotonicity matrix and complement according to the
        scipy template for LinearConstraints to use it
        in the minimization problem. Adding boundary condition.

        Returns
        -------
        linear_constraint_matrix: array-like of shape (n_linear_constraints, n_parameters)
            The matrix representing all linear constraints
            for needed parameters.
        """

        monotonicity_matrix = self.get_monotonicity_matrix()
        boundary_constraint = np.ones(self._get_number_of_moebius_coefficients())

        linear_constraint_matrix = np.concatenate(([boundary_constraint], monotonicity_matrix), axis=0)
        linear_constraint_matrix = np.insert(linear_constraint_matrix, 0, values=0, axis=1)
        linear_constraint_matrix = np.insert(linear_constraint_matrix, 0, values=0, axis=1)

        return linear_constraint_matrix

    def _get_number_of_moebius_coefficients(self):
        """Computes the number of needed moebius coefficients
        according to number of features and additivity.

        Returns
        -------
        number_of_moebius_coefficients: int
            Number of needed moebius coefficients for
            given problem
        """

        number_of_moebius_coefficients = 0

        for i in range(1, self.additivity + 1):
            number_of_moebius_coefficients += comb(self.number_of_features, i)

        return number_of_moebius_coefficients

    def _l1_regularization(self, moebius_coefficients):
        """Calculates the l1-Regularizer in the log-likelihood
        function to prevent overfitting.

        Parameters
        -------
        moebius_coefficients: array-like of shape (1, n_moebius_coefficients)
            Array containing all moebius coefficients

        Returns
        -------
        The calculated L1-Regularization for given problem
        """

        return self.regularization_parameter * sum(abs(coefficient) for coefficient in moebius_coefficients)



import numpy as np
import unittest

from classifier.mediator.components.feature_transformation import FeatureTransformation


class TestFeatureTransformation(unittest.TestCase):

    def test_no_redundant_data(self):
        data = np.array([[1, 2], [2, 1], [3, 4], [4, 3], [5, 5]])
        normalized_data = np.array([[0., 0.25], [0.25, 0.], [0.5, 0.75], [0.75, 0.5], [1., 1.]])

        f = FeatureTransformation(data)
        np.testing.assert_almost_equal(f.normalized, normalized_data, 5)

    def test_feature_trans(self):
        X = [[0.1, 0.2, 0.3], [0.2, 0.3, 0.4], [0.3, 0.4, 0.5], [0.4, 0.5, 0.6]]

        f = FeatureTransformation(X)

        instance = np.array([0.3, 0.3, 0.4])
        expected_instance = np.array([0.66666667, 0.33333333, 0.33333333])

        np.testing.assert_almost_equal([expected_instance], f([instance]))

    def test_redundant_data(self):
        X = np.array([[1, 2, 4], [1, 1, 2], [3, 3, 1], [3, 4, 2], [5, 1, 3]])
        normalized_X = np.array([[0., 0.5, 1], [0., 0., 0.375], [0.625, 0.75, 0], [0.625, 1, 0.375], [1., 0, 0.75]])

        f = FeatureTransformation(X)

        np.testing.assert_almost_equal(f.normalized, normalized_X, 5)


if __name__ == '__main__':
    unittest.main()




import numpy as np
import numpy.testing as nptest

from classifier.mediator.components import parameter_estimation as est
from classifier.mediator.components import helper as h
import unittest


class TestParameterEstimation(unittest.TestCase):
    def test_constraints_size(self):
        X = [[1, 2, 3, 5], [2, 3, 4, 1], [3, 4, 5, 2], [4, 5, 6, 6]]
        y = [0, 1]

        additivity = 4
        number_of_features = np.shape(X)[1]

        p = est.ParameterEstimation(X, y, additivity)
        number_of_moebius_coefficients = p._get_number_of_moebius_coefficients()

        # scaling (gamma) + threshold (beta) + moebius coefficients
        number_of_parameters = number_of_moebius_coefficients + 2

        number_of_monotonicity_constraints = np.power(2,
                                                      number_of_features - 1) * number_of_features - number_of_features
        number_of_linear_constraints = 1 + number_of_monotonicity_constraints

        bounds, linear_constraints = p._set_constraints()

        self.assertEqual(len(bounds.lb), number_of_parameters)
        self.assertEqual(len(bounds.ub), number_of_parameters)

        linear_constraint_matrix_columns = np.shape(linear_constraints.A)[1]
        linear_constraint_matrix_rows = np.shape(linear_constraints.A)[0]

        self.assertEqual(linear_constraint_matrix_columns, number_of_parameters)
        self.assertEqual(linear_constraint_matrix_rows, number_of_linear_constraints)

    def test_monotonicity_matrix(self):
        X = [[1, 2, 3], [2, 3, 4], [3, 4, 5], [4, 5, 6]]
        y = [0, 1]

        additivity = 3
        number_of_features = np.shape(X)[1]
        number_of_monotonicity_constraints = np.power(2,
                                                      number_of_features - 1) * number_of_features - number_of_features

        powersets = h.get_powerset_dictionary(list(range(1, number_of_features + 1)), additivity)
        p = est.ParameterEstimation(X, y, additivity, 1)
        number_of_moebius_coefficients = p._get_number_of_moebius_coefficients()
        monotonicity_matrix = p.get_monotonicity_matrix()

        self.assertEqual(number_of_monotonicity_constraints, np.shape(monotonicity_matrix)[0])

        monotonicity_sets_matrix = []
        for monotonicity_array in monotonicity_matrix:
            monotonicity_check_array = []
            moebius_coefficients_position = np.where(monotonicity_array == 1)[0]
            for index in moebius_coefficients_position:
                monotonicity_check_array.append(powersets[index + 1])
            monotonicity_sets_matrix.append(monotonicity_check_array)

        build_up_monotonicity_matrix = []
        for monotonicity_constraint in monotonicity_sets_matrix:
            arr = np.zeros(number_of_moebius_coefficients)
            indicator_array = [key - 1 for key, value in powersets.items() if value in monotonicity_constraint]
            monotonicity_matrix_row = [1 if index in indicator_array else 0 for index, x in enumerate(arr)]
            build_up_monotonicity_matrix.append(monotonicity_matrix_row)

        nptest.assert_array_equal(monotonicity_matrix, np.array(build_up_monotonicity_matrix))

    def test_linear_constraint_matrix(self):
        X = [[0.1, 0.2, 0.3], [0.2, 0.3, 0.4], [0.3, 0.4, 0.5], [0.4, 0.5, 0.6]]
        y = [1, 0, 1, 0]

        additivity = 3
        number_of_features = np.shape(X)[1]
        number_of_linear_constraints = np.power(2, number_of_features - 1) * number_of_features - \
                                       number_of_features + 1

        p = est.ParameterEstimation(X, y, additivity, 1)
        number_of_parameters = p._get_number_of_moebius_coefficients() + 2
        linear_constraint_matrix = p._get_linear_constraint_matrix()

        self.assertEqual(number_of_linear_constraints, np.shape(linear_constraint_matrix)[0])
        self.assertEqual(number_of_parameters, np.shape(linear_constraint_matrix)[1])

    def test_compute_parameters(self):
        X = [[0.1, 0.2, 0.3], [0.2, 0.3, 0.4], [0.3, 0.4, 0.5], [0.4, 0.5, 0.6]]
        y = [0, 0, 0, 1]

        p = est.ParameterEstimation(X, y, 2, 0.001)
        parameter_dict = p.compute_parameters()
        self.assertAlmostEqual(parameter_dict['beta'], 0.49458, places=5)

    def test_monotonicity_from_moebius_coefficients(self):
        X = [[0.1, 0.2, 0.3], [0.2, 0.3, 0.4], [0.3, 0.4, 0.5], [0.4, 0.5, 0.6]]
        y = [0, 0, 0, 1]

        additivity = 3
        number_of_features = np.shape(X)[1]
        number_of_monotonicity_constraints = np.power(2,
                                                      number_of_features - 1) * number_of_features - number_of_features

        p = est.ParameterEstimation(X, y, additivity, 1)
        parameter_dict = p.compute_parameters()

        moebius_coefficients = [frozenset({1}), frozenset({2}), frozenset({3}), frozenset({1, 2}),
                                frozenset({1, 3}), frozenset({2, 3}), frozenset({1, 2, 3})]
        moebius_dict = {key: parameter_dict[key] for key in moebius_coefficients}

        coefficient_values = list(moebius_dict.values())
        capacity_of_highest_subset = np.sum(coefficient_values)
        self.assertAlmostEqual(capacity_of_highest_subset, 1)

        capacity_c1 = coefficient_values[0]
        capacity_c2 = coefficient_values[1]
        capacity_c3 = coefficient_values[2]

        capacity_c1_c2 = capacity_c1 + capacity_c2 + coefficient_values[3]
        capacity_c1_c3 = capacity_c1 + capacity_c3 + coefficient_values[4]
        capacity_c2_c3 = capacity_c2 + capacity_c3 + coefficient_values[5]

        self.assertEqual((capacity_c1_c2 >= capacity_c1) & (capacity_c1_c2 >= capacity_c2), True)
        self.assertEqual((capacity_c1_c3 >= capacity_c1) & (capacity_c1_c3 >= capacity_c3), True)
        self.assertEqual((capacity_c2_c3 >= capacity_c2) & (capacity_c2_c3 >= capacity_c3), True)

    def test_l1_regularization(self):
        X = [[.25, .3, .7, .1], [.5, .625, .4, .75], [.35, .4, .5, .4], [.125, .5, .625, .6]]
        y = [1, 1, 1, 0]

        moebius_coefficient = [0.3, 0.3, 0.2, 0.2]
        p = est.ParameterEstimation(X, y, 1, 1)

        self.assertEqual(1.0, p._l1_regularization(moebius_coefficient))


if __name__ == '__main__':
    unittest.main()



import numpy as np

from .components.feature_transformation import FeatureTransformation
from .components.parameter_estimation import ParameterEstimation


class Fitter:
    """Fitter

    This class implements the functions to fit the feature transformation
    and the parameters of the Choquet classifier.

    """
    def __init__(self):
        pass

    def fit_feature_transformation(self, X):
        """ Fits the feature transformation for given input data.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        feature_transformation : FeatureTransformation
            Instance of the fitted feature transformation.
        """

        feature_transformation = FeatureTransformation(X)

        return feature_transformation

    def fit_parameters(self, X, y, additivity, regularization_parameter):
        """Fits the parameters threshold, scaling and moebius transform
        of the fuzzy measure for given input data.

        Parameters
        -------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like of shape (1, n_samples)
            Target labels to X.

        additivity : int or None.
            Additivity of fuzzy measure. If None, additivity will
            be set to 2 to estimate parameters.

        regularization_parameter : float or None
            the regularization parameter of the L1-Regularization.
            If None, regularization will be set to 1 to estimate the parameters.

        Returns
        -------
        parameter_dict : dict
            Dictionary, describing all fitted parameters (scaling, threshold,
            moebius coefficients) in the following manner:
            {'gamma': scaling, 'beta': threshold, frozenset({1}): m({1}),...}
        """

        if additivity is None:
            additivity = 2
        else:
            additivity = additivity

        if regularization_parameter is None:
            regularization_parameter = 0.001
        else:
            regularization_parameter = regularization_parameter

        parameter_estimation = ParameterEstimation(X, y, additivity, regularization_parameter)

        self.parameter_dict = parameter_estimation.compute_parameters()

        return self.parameter_dict

    def get_scaling_factor(self):
        return list(self.parameter_dict.values())[0]

    def get_threshold(self):
        return list(self.parameter_dict.values())[1]

    def get_moebius_transform_of_capacity(self):
        return dict(list(self.parameter_dict.items())[2:])








import unittest
from classifier.mediator.components.choquistic_regression import ChoquisticRegression


class TestChoquisticRegression(unittest.TestCase):
    def test_regression_calculation(self):
        X = [[0.25, 0.75, 0.25], [0.75, 0.25, 0.75]]
        y = [0, 1]
        parameter_dict = {'gamma': 82.4370341720885, 'beta': 0.48038295029897504, frozenset({1}): 0.32592547605321703,
                frozenset({2}): 0.016041218169457312, frozenset({3}): 0.4061951936952968,
                frozenset({1, 2}): 0.032105832146839916, frozenset({1, 3}): 0.1743782358631544,
                frozenset({2, 3}): 0.0317775846834257, frozenset({1, 2, 3}): 0.013576459388608858}
        regression = ChoquisticRegression(3, parameter_dict)

        self.assertAlmostEqual(regression.compute_regression_value(X[0]), y[0])
        self.assertAlmostEqual(regression.compute_regression_value(X[1]), y[1])


if __name__ == '__main__':
    unittest.main()



import numpy as np

from sklearn.preprocessing import QuantileTransformer


class FeatureTransformation:
    """Feature Transformation

        Using Quantile Transformation Scaler from scikit-learn


    Parameters
    ----------
    X : array-like of shape (n_samples, n_features)
        Input data, where n_samples is the number of samples and
        n_features is the number of features.
    """

    def __init__(self, X):
        self.normalized = self._initialize(X)

    def _initialize(self, X):
        """Initialize the function.
                Create an array of shape (1, n_features), where n_features is
                the number of features. Each index stores an instance of the
                class _FeatureTransformationComponent describing a component f_i
                Parameters:
                ----------
                X : array-like of shape (n_samples, n_features)
                    Input data, where n_samples is the number of samples and
                    n_features is the number of features.
                Returns:
                ----------
                array : ndarray of shape (, n_features)
                    Array containing a _FeatureTransformationComponent f_i for
                    each feature.
                """
        self.transformer = QuantileTransformer(n_quantiles=np.shape(X)[0], random_state=0)
        return self.transformer.fit_transform(X)

    def __getitem__(self, i):
        # use the item operator to access the feature transformation
        # component f_i
        return self.normalized[:, i]

    def __call__(self, x):
        return self.transformer.transform(x)

    #def __call__(self, x):
        # use the call operator to compute a normalized value for x



import unittest
from classifier.mediator.mediator import Mediator
from sklearn.utils.validation import check_array

class TestMediator(unittest.TestCase):

    def test_fit_components(self):
        # test for correct additivity
        X = [[0.1, 0.2, 0.3], [0.2, 0.3, 0.4], [0.3, 0.4, 0.5], [0.4, 0.5, 0.6]]
        y = [0, 0, 0, 1]

        additivity = 4

        mediator = Mediator()
        X = check_array(X)

        with self.assertRaises(ValueError):
            mediator.fit_components(X, y, additivity, 0.001)

    def test_for_regression_targets(self):
        # This check is copied from the Sugeno classifier by Sven Meyer:
        # https://github.com/smeyer198/sugeno-classifier/blob/main/classifier/tests/mediator/test_mediator.py
        X = [[1, 2, 3], [2, 3, 4], [4, 5, 6]]

        mediator = Mediator()

        # no regression targets
        y = [2, 5, 2]

        result = mediator._check_for_regression_targets(y)

        self.assertTrue(result)

        # regression targets
        y = [0.1, 0.4, 0.1]

        result = mediator._check_for_regression_targets(y)

        self.assertFalse(result)

        # mixed targets
        y = [3, 6, 0.5]

        result = mediator._check_for_regression_targets(y)

        self.assertFalse(result)

        # ints are allowed to be a floating point
        y = [3.0, 6.0, 5.0]

        result = mediator._check_for_regression_targets(y)

        self.assertTrue(result)

        # strings are considered to be labels and no regression targets
        y = [3, '3.7', 1]

        result = mediator._check_for_regression_targets(y)

        self.assertTrue(result)



if __name__ == '__main__':
    unittest.main()
